/*-
 * Copyright 2009 Colin Percival
 * Copyright 2012-2014 Alexander Peslyak
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * This file was originally written by Colin Percival as part of the Tarsnap
 * online backup system.
 */

#include <emmintrin.h>
#ifdef __XOP__
#include <x86intrin.h>
#endif

#include <errno.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#include "sysendian.h"
#include "sha256.h"

#include "crypto_scrypt.h"

#include "scrypt_platform.c"

#if __STDC_VERSION__ >= 199901L
/* have restrict */
#elif defined(__GNUC__)
#define restrict __restrict
#else
#define restrict
#endif

#define PREFETCH(x, hint) _mm_prefetch((const char *)(x), (hint));
#define PREFETCH_OUT(x, hint) /* disabled */

#ifdef __XOP__
#define ARX(out, in1, in2, s) \
	out = _mm_xor_si128(out, _mm_roti_epi32(_mm_add_epi32(in1, in2), s));
#else
#define ARX(out, in1, in2, s) \
	{ \
		__m128i T = _mm_add_epi32(in1, in2); \
		out = _mm_xor_si128(out, _mm_slli_epi32(T, s)); \
		out = _mm_xor_si128(out, _mm_srli_epi32(T, 32-s)); \
	}
#endif

#define SALSA20_2ROUNDS \
	/* Operate on "columns". */ \
	ARX(X1, X0, X3, 7) \
	ARX(X2, X1, X0, 9) \
	ARX(X3, X2, X1, 13) \
	ARX(X0, X3, X2, 18) \
\
	/* Rearrange data. */ \
	X1 = _mm_shuffle_epi32(X1, 0x93); \
	X2 = _mm_shuffle_epi32(X2, 0x4E); \
	X3 = _mm_shuffle_epi32(X3, 0x39); \
\
	/* Operate on "rows". */ \
	ARX(X3, X0, X1, 7) \
	ARX(X2, X3, X0, 9) \
	ARX(X1, X2, X3, 13) \
	ARX(X0, X1, X2, 18) \
\
	/* Rearrange data. */ \
	X1 = _mm_shuffle_epi32(X1, 0x39); \
	X2 = _mm_shuffle_epi32(X2, 0x4E); \
	X3 = _mm_shuffle_epi32(X3, 0x93);

/**
 * Apply the salsa20/8 core to the block provided in (X0 ... X3).
 */
#define SALSA20_8_BASE(maybe_decl, out) \
	{ \
		maybe_decl Y0 = X0; \
		maybe_decl Y1 = X1; \
		maybe_decl Y2 = X2; \
		maybe_decl Y3 = X3; \
		SALSA20_2ROUNDS \
		SALSA20_2ROUNDS \
		SALSA20_2ROUNDS \
		SALSA20_2ROUNDS \
		(out)[0] = X0 = _mm_add_epi32(X0, Y0); \
		(out)[1] = X1 = _mm_add_epi32(X1, Y1); \
		(out)[2] = X2 = _mm_add_epi32(X2, Y2); \
		(out)[3] = X3 = _mm_add_epi32(X3, Y3); \
	}
#define SALSA20_8(out) \
	SALSA20_8_BASE(__m128i, out)

/**
 * Apply the salsa20/8 core to the block provided in (X0 ... X3) ^ (Z0 ... Z3).
 */
#define SALSA20_8_XOR_ANY(maybe_decl, Z0, Z1, Z2, Z3, out) \
	X0 = _mm_xor_si128(X0, Z0); \
	X1 = _mm_xor_si128(X1, Z1); \
	X2 = _mm_xor_si128(X2, Z2); \
	X3 = _mm_xor_si128(X3, Z3); \
	SALSA20_8_BASE(maybe_decl, out)

#define SALSA20_8_XOR_MEM(in, out) \
	SALSA20_8_XOR_ANY(__m128i, (in)[0], (in)[1], (in)[2], (in)[3], out)

#define SALSA20_8_XOR_REG(out) \
	SALSA20_8_XOR_ANY(/* empty */, Y0, Y1, Y2, Y3, out)

/**
 * blockmix_salsa8(Bin, Bout, r):
 * Compute Bout = BlockMix_{salsa20/8, r}(Bin).  The input Bin must be 128r
 * bytes in length; the output Bout must also be the same size.
 */
static inline void
blockmix_salsa8(const __m128i *restrict Bin, __m128i *restrict Bout, size_t r)
{
	__m128i X0, X1, X2, X3;
	size_t r4, r4p4, r8, r8p4;
	size_t i;

	r--;
	r4 = r * 4;
	r4p4 = r4 + 4;
	r8 = r * 8;
	r8p4 = r8 + 4;
	PREFETCH(&Bin[r8p4], _MM_HINT_T0)
	for (i = 0; i < r; i++) {
		PREFETCH(&Bin[i * 8], _MM_HINT_T0)
		PREFETCH_OUT(&Bout[i * 4], _MM_HINT_T0)
		PREFETCH(&Bin[i * 8 + 4], _MM_HINT_T0)
		PREFETCH_OUT(&Bout[i * 4 + r4p4], _MM_HINT_T0)
	}
	PREFETCH(&Bin[r8], _MM_HINT_T0)
	PREFETCH_OUT(&Bout[r4], _MM_HINT_T0)
	PREFETCH_OUT(&Bout[r8p4], _MM_HINT_T0)

	/* 1: X <-- B_{2r - 1} */
	X0 = Bin[r8p4];
	X1 = Bin[r8p4 + 1];
	X2 = Bin[r8p4 + 2];
	X3 = Bin[r8p4 + 3];

	/* 3: X <-- H(X \xor B_i) */
	/* 4: Y_i <-- X */
	/* 6: B' <-- (Y_0, Y_2 ... Y_{2r-2}, Y_1, Y_3 ... Y_{2r-1}) */
	SALSA20_8_XOR_MEM(Bin, Bout)

	/* 2: for i = 0 to 2r - 1 do */
	for (i = 0; i < r;) {
		/* 3: X <-- H(X \xor B_i) */
		/* 4: Y_i <-- X */
		/* 6: B' <-- (Y_0, Y_2 ... Y_{2r-2}, Y_1, Y_3 ... Y_{2r-1}) */
		SALSA20_8_XOR_MEM(&Bin[i * 8 + 4], &Bout[i * 4 + r4p4])

		i++;

		/* 3: X <-- H(X \xor B_i) */
		/* 4: Y_i <-- X */
		/* 6: B' <-- (Y_0, Y_2 ... Y_{2r-2}, Y_1, Y_3 ... Y_{2r-1}) */
		SALSA20_8_XOR_MEM(&Bin[i * 8], &Bout[i * 4])
	}

	/* 3: X <-- H(X \xor B_i) */
	/* 4: Y_i <-- X */
	/* 6: B' <-- (Y_0, Y_2 ... Y_{2r-2}, Y_1, Y_3 ... Y_{2r-1}) */
	SALSA20_8_XOR_MEM(&Bin[r8p4], &Bout[r8p4])
}

#define S_SIZE_ALL (2 * 0x100 * 0x10)

#if defined(__x86_64__) && (defined(__ICC) || defined(__llvm__))
/* Intel's name, also supported by recent gcc */
#define EXTRACT64(X) _mm_cvtsi128_si64(X)
#elif defined(__x86_64__) && !defined(_MSC_VER) && !defined(__OPEN64__)
/* gcc got the 'x' name earlier than non-'x', MSVC and Open64 had bugs */
#define EXTRACT64(X) _mm_cvtsi128_si64x(X)
#elif defined(__x86_64__) && defined(__SSE4_1__)
/* No known bugs for this intrinsic */
#include <smmintrin.h>
#define EXTRACT64(X) _mm_extract_epi64((X), 0)
#elif defined(__SSE4_1__)
/* 32-bit */
#include <smmintrin.h>
#define EXTRACT64(X) \
	((uint64_t)(uint32_t)_mm_cvtsi128_si32(X) | \
	((uint64_t)(uint32_t)_mm_extract_epi32((X), 1) << 32))
#else
/* 32-bit or compilers with known past bugs in _mm_cvtsi128_si64*() */
#define EXTRACT64(X) \
	((uint64_t)(uint32_t)_mm_cvtsi128_si32(X) | \
	((uint64_t)(uint32_t)_mm_cvtsi128_si32(_mm_srli_epi64((X), 32)) << 32))
#endif

#define SBOX_SIMD_1(X, x, s0, s1) \
	x = EXTRACT64(X) & 0x0ff000000ff0ULL; \
	s0 = *(const __m128i *)(S0 + (uint32_t)x); \
	s1 = *(const __m128i *)(S1 + (x >> 32)); \

#define SBOX_SIMD_2(X, x, s0, s1) \
	X = _mm_add_epi64(X, _mm_xor_si128( \
	    _mm_add_epi64(_mm_mul_epu32(X, s0), s1), \
	    _mm_add_epi64(_mm_mul_epu32(X, s1), s0)));

#define SBOX_ROUND \
	SBOX_SIMD_1(X0, x0, s00, s01) \
	SBOX_SIMD_1(X1, x1, s10, s11) \
	SBOX_SIMD_2(X0, x0, s00, s01) \
	SBOX_SIMD_2(X1, x1, s10, s11) \
	SBOX_SIMD_1(X2, x2, s20, s21) \
	SBOX_SIMD_1(X3, x3, s30, s31) \
	SBOX_SIMD_2(X2, x2, s20, s21) \
	SBOX_SIMD_2(X3, x3, s30, s31)

#define SBOX \
	{ \
		uint64_t x0, x1, x2, x3; \
		__m128i s00, s01, s10, s11, s20, s21, s30, s31; \
		SBOX_ROUND SBOX_ROUND \
		SBOX_ROUND SBOX_ROUND \
	}

#define XOR4(in) \
	X0 = _mm_xor_si128(X0, (in)[0]); \
	X1 = _mm_xor_si128(X1, (in)[1]); \
	X2 = _mm_xor_si128(X2, (in)[2]); \
	X3 = _mm_xor_si128(X3, (in)[3]);

#define OUT(out) \
	(out)[0] = X0; \
	(out)[1] = X1; \
	(out)[2] = X2; \
	(out)[3] = X3;

/**
 * blockmix_sbox(Bin, Bout, S, r):
 * Compute Bout = BlockMix_Sbox{salsa20/8, S, r}(Bin).  The input Bin must be
 * 128r bytes in length; the output Bout must also be the same size.
 */
static inline void
blockmix_sbox(const __m128i *restrict Bin, __m128i *restrict Bout,
    const __m128i *restrict S, size_t r)
{
	const uint8_t * S0 = (const uint8_t *)S;
	const uint8_t * S1 = (const uint8_t *)S + S_SIZE_ALL / 2;
	__m128i X0, X1, X2, X3;
	size_t r4;
	size_t i;

	/* Convert 128-byte blocks to 16-byte blocks */
	r4 = r * 8;
	/* Convert 128-byte blocks to 64-byte blocks */
	r *= 2;

	PREFETCH(&Bin[r4 - 4], _MM_HINT_T0)
	r--;
	for (i = 0; i < r; i++) {
		PREFETCH(&Bin[i * 4], _MM_HINT_T0)
		PREFETCH_OUT(&Bout[i * 4], _MM_HINT_T0)
	}
	PREFETCH_OUT(&Bout[r4 - 4], _MM_HINT_T0)

	/* X <-- B_{r1 - 1} */
	X0 = Bin[r4 - 4];
	X1 = Bin[r4 - 3];
	X2 = Bin[r4 - 2];
	X3 = Bin[r4 - 1];

	/* for i = 0 to r1 - 1 do */
	for (i = 0; i < r; i++) {
		/* X <-- H'(X \xor B_i) */
		XOR4(&Bin[i * 4])
		SBOX
		/* B'_i <-- X */
		OUT(&Bout[i * 4])
	}

	/* Last iteration of the loop above */
	XOR4(&Bin[i * 4])
	SBOX

	/* B'_i <-- H(B'_i) */
	SALSA20_8(&Bout[i * 4])
}

#define XOR4_2(in1, in2) \
	X0 = _mm_xor_si128((in1)[0], (in2)[0]); \
	X1 = _mm_xor_si128((in1)[1], (in2)[1]); \
	X2 = _mm_xor_si128((in1)[2], (in2)[2]); \
	X3 = _mm_xor_si128((in1)[3], (in2)[3]);

static inline uint32_t
blockmix_salsa8_xor(const __m128i *restrict Bin1, const __m128i *restrict Bin2,
    __m128i *restrict Bout, size_t r, int Bin2_in_ROM)
{
	__m128i X0, X1, X2, X3;
	size_t r4, r4p4, r8, r8p4;
	size_t i;

	r--;
	r4 = r * 4;
	r4p4 = r4 + 4;
	r8 = r * 8;
	r8p4 = r8 + 4;
	if (Bin2_in_ROM) {
		PREFETCH(&Bin2[r8p4], _MM_HINT_NTA)
		PREFETCH(&Bin1[r8p4], _MM_HINT_T0)
		for (i = 0; i < r; i++) {
			PREFETCH(&Bin2[i * 8], _MM_HINT_NTA)
			PREFETCH(&Bin1[i * 8], _MM_HINT_T0)
			PREFETCH(&Bin2[i * 8 + 4], _MM_HINT_NTA)
			PREFETCH(&Bin1[i * 8 + 4], _MM_HINT_T0)
			PREFETCH_OUT(&Bout[i * 4], _MM_HINT_T0)
			PREFETCH_OUT(&Bout[i * 4 + r4p4], _MM_HINT_T0)
		}
		PREFETCH(&Bin2[r8], _MM_HINT_NTA)
	} else {
		PREFETCH(&Bin2[r8p4], _MM_HINT_T0)
		PREFETCH(&Bin1[r8p4], _MM_HINT_T0)
		for (i = 0; i < r; i++) {
			PREFETCH(&Bin2[i * 8], _MM_HINT_T0)
			PREFETCH(&Bin1[i * 8], _MM_HINT_T0)
			PREFETCH(&Bin2[i * 8 + 4], _MM_HINT_T0)
			PREFETCH(&Bin1[i * 8 + 4], _MM_HINT_T0)
			PREFETCH_OUT(&Bout[i * 4], _MM_HINT_T0)
			PREFETCH_OUT(&Bout[i * 4 + r4p4], _MM_HINT_T0)
		}
		PREFETCH(&Bin2[r8], _MM_HINT_T0)
	}
	PREFETCH(&Bin1[r8], _MM_HINT_T0)
	PREFETCH_OUT(&Bout[r4], _MM_HINT_T0)
	PREFETCH_OUT(&Bout[r8p4], _MM_HINT_T0)

	/* 1: X <-- B_{2r - 1} */
	XOR4_2(&Bin1[r8p4], &Bin2[r8p4])

	/* 3: X <-- H(X \xor B_i) */
	/* 4: Y_i <-- X */
	/* 6: B' <-- (Y_0, Y_2 ... Y_{2r-2}, Y_1, Y_3 ... Y_{2r-1}) */
	XOR4(Bin1)
	SALSA20_8_XOR_MEM(Bin2, Bout)

	/* 2: for i = 0 to 2r - 1 do */
	for (i = 0; i < r;) {
		/* 3: X <-- H(X \xor B_i) */
		/* 4: Y_i <-- X */
		/* 6: B' <-- (Y_0, Y_2 ... Y_{2r-2}, Y_1, Y_3 ... Y_{2r-1}) */
		XOR4(&Bin1[i * 8 + 4])
		SALSA20_8_XOR_MEM(&Bin2[i * 8 + 4], &Bout[i * 4 + r4p4])

		i++;

		/* 3: X <-- H(X \xor B_i) */
		/* 4: Y_i <-- X */
		/* 6: B' <-- (Y_0, Y_2 ... Y_{2r-2}, Y_1, Y_3 ... Y_{2r-1}) */
		XOR4(&Bin1[i * 8])
		SALSA20_8_XOR_MEM(&Bin2[i * 8], &Bout[i * 4])
	}

	/* 3: X <-- H(X \xor B_i) */
	/* 4: Y_i <-- X */
	/* 6: B' <-- (Y_0, Y_2 ... Y_{2r-2}, Y_1, Y_3 ... Y_{2r-1}) */
	XOR4(&Bin1[r8p4])
	SALSA20_8_XOR_MEM(&Bin2[r8p4], &Bout[r8p4])

	return _mm_cvtsi128_si32(X0);
}

static inline uint32_t
blockmix_sbox_xor(const __m128i *restrict Bin1, const __m128i *restrict Bin2,
    __m128i *restrict Bout, const __m128i *restrict S, size_t r,
    int Bin2_in_ROM)
{
	const uint8_t * S0 = (const uint8_t *)S;
	const uint8_t * S1 = (const uint8_t *)S + S_SIZE_ALL / 2;
	__m128i X0, X1, X2, X3;
	size_t r4;
	size_t i;

	/* Convert 128-byte blocks to 16-byte blocks */
	r4 = r * 8;
	/* Convert 128-byte blocks to 64-byte blocks */
	r *= 2;

	r--;
	if (Bin2_in_ROM) {
		PREFETCH(&Bin2[r4 - 4], _MM_HINT_NTA)
		PREFETCH(&Bin1[r4 - 4], _MM_HINT_T0)
		for (i = 0; i < r; i++) {
			PREFETCH(&Bin2[i * 4], _MM_HINT_NTA)
			PREFETCH(&Bin1[i * 4], _MM_HINT_T0)
			PREFETCH_OUT(&Bout[i * 4], _MM_HINT_T0)
		}
	} else {
		PREFETCH(&Bin2[r4 - 4], _MM_HINT_T0)
		PREFETCH(&Bin1[r4 - 4], _MM_HINT_T0)
		for (i = 0; i < r; i++) {
			PREFETCH(&Bin2[i * 4], _MM_HINT_T0)
			PREFETCH(&Bin1[i * 4], _MM_HINT_T0)
			PREFETCH_OUT(&Bout[i * 4], _MM_HINT_T0)
		}
	}
	PREFETCH_OUT(&Bout[r4 - 4], _MM_HINT_T0);

	/* X <-- B_{r1 - 1} */
	XOR4_2(&Bin1[r4 - 4], &Bin2[r4 - 4])

	/* for i = 0 to r1 - 1 do */
	for (i = 0; i < r; i++) {
		/* X <-- H'(X \xor B_i) */
		XOR4(&Bin1[i * 4])
		XOR4(&Bin2[i * 4])
		SBOX
		/* B'_i <-- X */
		OUT(&Bout[i * 4])
	}

	/* Last iteration of the loop above */
	XOR4(&Bin1[i * 4])
	XOR4(&Bin2[i * 4])
	SBOX

	/* B'_i <-- H(B'_i) */
	SALSA20_8(&Bout[i * 4])

	return _mm_cvtsi128_si32(X0);
}

#undef XOR4
#define XOR4(in, out) \
	(out)[0] = Y0 = _mm_xor_si128((in)[0], (out)[0]); \
	(out)[1] = Y1 = _mm_xor_si128((in)[1], (out)[1]); \
	(out)[2] = Y2 = _mm_xor_si128((in)[2], (out)[2]); \
	(out)[3] = Y3 = _mm_xor_si128((in)[3], (out)[3]);

static inline uint32_t
blockmix_salsa8_xor_save(const __m128i *restrict Bin1, __m128i *restrict Bin2,
    __m128i *restrict Bout, size_t r)
{
	__m128i X0, X1, X2, X3, Y0, Y1, Y2, Y3;
	size_t r4, r4p4, r8, r8p4;
	size_t i;

	r--;
	r4 = r * 4;
	r4p4 = r4 + 4;
	r8 = r * 8;
	r8p4 = r8 + 4;
	PREFETCH(&Bin2[r8p4], _MM_HINT_T0)
	PREFETCH(&Bin1[r8p4], _MM_HINT_T0)
	for (i = 0; i < r; i++) {
		PREFETCH(&Bin2[i * 8], _MM_HINT_T0)
		PREFETCH(&Bin1[i * 8], _MM_HINT_T0)
		PREFETCH(&Bin2[i * 8 + 4], _MM_HINT_T0)
		PREFETCH(&Bin1[i * 8 + 4], _MM_HINT_T0)
		PREFETCH_OUT(&Bout[i * 4], _MM_HINT_T0)
		PREFETCH_OUT(&Bout[i * 4 + r4p4], _MM_HINT_T0)
	}
	PREFETCH(&Bin2[r8], _MM_HINT_T0)
	PREFETCH(&Bin1[r8], _MM_HINT_T0)
	PREFETCH_OUT(&Bout[r4], _MM_HINT_T0);
	PREFETCH_OUT(&Bout[r8p4], _MM_HINT_T0);

	/* 1: X <-- B_{2r - 1} */
	XOR4_2(&Bin1[r8p4], &Bin2[r8p4])

	/* 3: X <-- H(X \xor B_i) */
	/* 4: Y_i <-- X */
	/* 6: B' <-- (Y_0, Y_2 ... Y_{2r-2}, Y_1, Y_3 ... Y_{2r-1}) */
	XOR4(Bin1, Bin2)
	SALSA20_8_XOR_REG(Bout)

	/* 2: for i = 0 to 2r - 1 do */
	for (i = 0; i < r;) {
		/* 3: X <-- H(X \xor B_i) */
		/* 4: Y_i <-- X */
		/* 6: B' <-- (Y_0, Y_2 ... Y_{2r-2}, Y_1, Y_3 ... Y_{2r-1}) */
		XOR4(&Bin1[i * 8 + 4], &Bin2[i * 8 + 4])
		SALSA20_8_XOR_REG(&Bout[i * 4 + r4p4])

		i++;

		/* 3: X <-- H(X \xor B_i) */
		/* 4: Y_i <-- X */
		/* 6: B' <-- (Y_0, Y_2 ... Y_{2r-2}, Y_1, Y_3 ... Y_{2r-1}) */
		XOR4(&Bin1[i * 8], &Bin2[i * 8])
		SALSA20_8_XOR_REG(&Bout[i * 4])
	}

	/* 3: X <-- H(X \xor B_i) */
	/* 4: Y_i <-- X */
	/* 6: B' <-- (Y_0, Y_2 ... Y_{2r-2}, Y_1, Y_3 ... Y_{2r-1}) */
	XOR4(&Bin1[r8p4], &Bin2[r8p4])
	SALSA20_8_XOR_REG(&Bout[r8p4])

	return _mm_cvtsi128_si32(X0);
}

#define XOR4_Y \
	X0 = _mm_xor_si128(X0, Y0); \
	X1 = _mm_xor_si128(X1, Y1); \
	X2 = _mm_xor_si128(X2, Y2); \
	X3 = _mm_xor_si128(X3, Y3);

static inline uint32_t
blockmix_sbox_xor_save(const __m128i *restrict Bin1, __m128i *restrict Bin2,
    __m128i *restrict Bout, const __m128i *restrict S, size_t r)
{
	const uint8_t * S0 = (const uint8_t *)S;
	const uint8_t * S1 = (const uint8_t *)S + S_SIZE_ALL / 2;
	__m128i X0, X1, X2, X3, Y0, Y1, Y2, Y3;
	size_t r4;
	size_t i;

	/* Convert 128-byte blocks to 16-byte blocks */
	r4 = r * 8;
	/* Convert 128-byte blocks to 64-byte blocks */
	r *= 2;

	r--;
	PREFETCH(&Bin2[r4 - 4], _MM_HINT_T0)
	PREFETCH(&Bin1[r4 - 4], _MM_HINT_T0)
	for (i = 0; i < r; i++) {
		PREFETCH(&Bin2[i * 4], _MM_HINT_T0)
		PREFETCH(&Bin1[i * 4], _MM_HINT_T0)
		PREFETCH_OUT(&Bout[i * 4], _MM_HINT_T0)
	}
	PREFETCH_OUT(&Bout[r4 - 4], _MM_HINT_T0)

	/* X <-- B_{r1 - 1} */
	XOR4_2(&Bin1[r4 - 4], &Bin2[r4 - 4])

	/* for i = 0 to r1 - 1 do */
	for (i = 0; i < r; i++) {
		XOR4(&Bin1[i * 4], &Bin2[i * 4])
		/* X <-- H'(X \xor B_i) */
		XOR4_Y
		SBOX
		/* B'_i <-- X */
		OUT(&Bout[i * 4])
	}

	/* Last iteration of the loop above */
	XOR4(&Bin1[i * 4], &Bin2[i * 4])
	XOR4_Y
	SBOX

	/* B'_i <-- H(B'_i) */
	SALSA20_8(&Bout[i * 4])

	return _mm_cvtsi128_si32(X0);
}

#undef ARX
#undef SALSA20_2ROUNDS
#undef SALSA20_8
#undef SALSA20_8_XOR_ANY
#undef SALSA20_8_XOR_MEM
#undef SALSA20_8_XOR_REG
#undef SBOX_SIMD_1
#undef SBOX_SIMD_2
#undef SBOX_ROUND
#undef SBOX
#undef OUT
#undef XOR4
#undef XOR4_2
#undef XOR4_Y

/**
 * integerify(B, r):
 * Return the result of parsing B_{2r-1} as a little-endian integer.
 */
static inline uint32_t
integerify(const void * B, size_t r)
{
	return *(const uint32_t *)((uintptr_t)(B) + (2 * r - 1) * 64);
}

/**
 * smix1(B, r, N, flags, V, NROM, shared, XY, S):
 * Compute first loop of B = SMix_r(B, N).  The input B must be 128r bytes in
 * length; the temporary storage V must be 128rN bytes in length; the temporary
 * storage XY must be 128r bytes in length.  The value N must be even and no
 * smaller than 2.  The array V must be aligned to a multiple of 64 bytes, and
 * arrays B and XY to a multiple of at least 16 bytes (aligning them to 64
 * bytes as well saves cache lines, but might result in cache bank conflicts).
 */
static void
smix1(uint8_t * B, size_t r, uint32_t N, escrypt_flags_t flags, void * V,
    uint32_t NROM, const escrypt_shared_t * shared, void * XY, void * S)
{
	const void * VROM = shared->shared1.aligned;
	uint32_t VROM_mask = shared->mask1;
	size_t s = 128 * r;
	__m128i * X = V, * Y;
	uint32_t * X32 = V;
	uint32_t i, j;
	size_t k;

	/* 1: X <-- B */
	/* 3: V_i <-- X */
	for (k = 0; k < 2 * r; k++) {
		for (i = 0; i < 16; i++) {
			X32[k * 16 + i] =
			    le32dec(&B[(k * 16 + (i * 5 % 16)) * 4]);
		}
	}

	if (NROM && (VROM_mask & 1)) {
		uint32_t n;
		void * V_n;
		const __m128i * V_j;

		/* 4: X <-- H(X) */
		/* 3: V_i <-- X */
		Y = (void *)((uintptr_t)(V) + s);
		if (S)
			blockmix_sbox(X, Y, S, r);
		else
			blockmix_salsa8(X, Y, r);

		X = (void *)((uintptr_t)(V) + 2 * s);
		if ((1 & VROM_mask) == 1) {
			/* j <-- Integerify(X) mod NROM */
			j = integerify(Y, r) & (NROM - 1);
			V_j = (const void *)((uintptr_t)(VROM) + j * s);

			/* X <-- H(X \xor VROM_j) */
			if (S)
				j = blockmix_sbox_xor(Y, V_j, X, S, r, 1);
			else
				j = blockmix_salsa8_xor(Y, V_j, X, r, 1);
		} else {
			/* X <-- H(X) */
			if (S)
				blockmix_sbox(Y, X, S, r);
			else
				blockmix_salsa8(Y, X, r);
			j = integerify(X, r);
		}

		for (n = 2; n < N; n <<= 1) {
			uint32_t m = (n < N / 2) ? n : (N - 1 - n);

			V_n = (void *)((uintptr_t)(V) + n * s);

			/* 2: for i = 0 to N - 1 do */
			for (i = 1; i < m; i += 2) {
				/* j <-- Wrap(Integerify(X), i) */
				j &= n - 1;
				j += i - 1;
				V_j = (const void *)((uintptr_t)(V) + j * s);

				/* X <-- X \xor V_j */
				/* 4: X <-- H(X) */
				/* 3: V_i <-- X */
				Y = (void *)((uintptr_t)(V_n) + i * s);
				if (S)
					j = blockmix_sbox_xor(X, V_j, Y, S, r, 0);
				else
					j = blockmix_salsa8_xor(X, V_j, Y, r, 0);

				if (((n + i) & VROM_mask) == 1) {
					/* j <-- Integerify(X) mod NROM */
					j &= NROM - 1;
					V_j = (const void *)((uintptr_t)(VROM) + j * s);
				} else {
					/* j <-- Wrap(Integerify(X), i) */
					j &= n - 1;
					j += i;
					V_j = (const void *)((uintptr_t)(V) + j * s);
				}

				/* X <-- H(X \xor VROM_j) */
				X = (void *)((uintptr_t)(V_n) + (i + 1) * s);
				if (S)
					j = blockmix_sbox_xor(Y, V_j, X, S, r, 1);
				else
					j = blockmix_salsa8_xor(Y, V_j, X, r, 1);
			}
		}

		n >>= 1;

		/* j <-- Wrap(Integerify(X), i) */
		j &= n - 1;
		j += N - 2 - n;
		V_j = (const void *)((uintptr_t)(V) + j * s);

		/* X <-- X \xor V_j */
		/* 4: X <-- H(X) */
		/* 3: V_i <-- X */
		Y = (void *)((uintptr_t)(V) + (N - 1) * s);
		if (S)
			j = blockmix_sbox_xor(X, V_j, Y, S, r, 0);
		else
			j = blockmix_salsa8_xor(X, V_j, Y, r, 0);

		if (((N - 1) & VROM_mask) == 1) {
			/* j <-- Integerify(X) mod NROM */
			j &= NROM - 1;
			V_j = (const void *)((uintptr_t)(VROM) + j * s);
		} else {
			/* j <-- Wrap(Integerify(X), i) */
			j &= n - 1;
			j += N - 1 - n;
			V_j = (const void *)((uintptr_t)(V) + j * s);
		}

		/* X <-- X \xor V_j */
		/* 4: X <-- H(X) */
		X = XY;
		if (S)
			blockmix_sbox_xor(Y, V_j, X, S, r, 1);
		else
			blockmix_salsa8_xor(Y, V_j, X, r, 1);
	} else if (flags & ESCRYPT_RW) {
		uint32_t n;
		void * V_n;
		const __m128i * V_j;

		/* 4: X <-- H(X) */
		/* 3: V_i <-- X */
		Y = (void *)((uintptr_t)(V) + s);
		if (S)
			blockmix_sbox(X, Y, S, r);
		else
			blockmix_salsa8(X, Y, r);

		/* 4: X <-- H(X) */
		/* 3: V_i <-- X */
		X = (void *)((uintptr_t)(V) + 2 * s);
		if (S)
			blockmix_sbox(Y, X, S, r);
		else
			blockmix_salsa8(Y, X, r);
		j = integerify(X, r);

		for (n = 2; n < N; n <<= 1) {
			uint32_t m = (n < N / 2) ? n : (N - 1 - n);

			V_n = (void *)((uintptr_t)(V) + n * s);

			/* 2: for i = 0 to N - 1 do */
			for (i = 1; i < m; i += 2) {
				Y = (void *)((uintptr_t)(V_n) + i * s);

				/* j <-- Wrap(Integerify(X), i) */
				j &= n - 1;
				j += i - 1;
				V_j = (const void *)((uintptr_t)(V) + j * s);

				/* X <-- X \xor V_j */
				/* 4: X <-- H(X) */
				/* 3: V_i <-- X */
				if (S)
					j = blockmix_sbox_xor(X, V_j, Y, S, r, 0);
				else
					j = blockmix_salsa8_xor(X, V_j, Y, r, 0);

				/* j <-- Wrap(Integerify(X), i) */
				j &= n - 1;
				j += i;
				V_j = (const void *)((uintptr_t)(V) + j * s);

				/* X <-- X \xor V_j */
				/* 4: X <-- H(X) */
				/* 3: V_i <-- X */
				X = (void *)((uintptr_t)(V_n) + (i + 1) * s);
				if (S)
					j = blockmix_sbox_xor(Y, V_j, X, S, r, 0);
				else
					j = blockmix_salsa8_xor(Y, V_j, X, r, 0);
			}
		}

		n >>= 1;

		/* j <-- Wrap(Integerify(X), i) */
		j &= n - 1;
		j += N - 2 - n;
		V_j = (const void *)((uintptr_t)(V) + j * s);

		/* X <-- X \xor V_j */
		/* 4: X <-- H(X) */
		/* 3: V_i <-- X */
		Y = (void *)((uintptr_t)(V) + (N - 1) * s);
		if (S)
			j = blockmix_sbox_xor(X, V_j, Y, S, r, 0);
		else
			j = blockmix_salsa8_xor(X, V_j, Y, r, 0);

		/* j <-- Wrap(Integerify(X), i) */
		j &= n - 1;
		j += N - 1 - n;
		V_j = (const void *)((uintptr_t)(V) + j * s);

		/* X <-- X \xor V_j */
		/* 4: X <-- H(X) */
		X = XY;
		if (S)
			blockmix_sbox_xor(Y, V_j, X, S, r, 0);
		else
			blockmix_salsa8_xor(Y, V_j, X, r, 0);
	} else {
		/* 2: for i = 0 to N - 1 do */
		for (i = 1; i < N - 1; i += 2) {
			/* 4: X <-- H(X) */
			/* 3: V_i <-- X */
			Y = (void *)((uintptr_t)(V) + i * s);
			if (S)
				blockmix_sbox(X, Y, S, r);
			else
				blockmix_salsa8(X, Y, r);

			/* 4: X <-- H(X) */
			/* 3: V_i <-- X */
			X = (void *)((uintptr_t)(V) + (i + 1) * s);
			if (S)
				blockmix_sbox(Y, X, S, r);
			else
				blockmix_salsa8(Y, X, r);
		}

		/* 4: X <-- H(X) */
		/* 3: V_i <-- X */
		Y = (void *)((uintptr_t)(V) + i * s);
		if (S)
			blockmix_sbox(X, Y, S, r);
		else
			blockmix_salsa8(X, Y, r);

		/* 4: X <-- H(X) */
		X = XY;
		if (S)
			blockmix_sbox(Y, X, S, r);
		else
			blockmix_salsa8(Y, X, r);
	}

	/* B' <-- X */
	X32 = XY;
	for (k = 0; k < 2 * r; k++) {
		for (i = 0; i < 16; i++) {
			le32enc(&B[(k * 16 + (i * 5 % 16)) * 4],
			    X32[k * 16 + i]);
		}
	}
}

/**
 * smix2(B, r, N, Nloop, flags, V, NROM, shared, XY, S):
 * Compute second loop of B = SMix_r(B, N).  The input B must be 128r bytes in
 * length; the temporary storage V must be 128rN bytes in length; the temporary
 * storage XY must be 256r bytes in length.  The value N must be a power of 2
 * greater than 1.  The value Nloop must be even.  The array V must be aligned
 * to a multiple of 64 bytes, and arrays B and XY to a multiple of at least 16
 * bytes (aligning them to 64 bytes as well saves cache lines, but might result
 * in cache bank conflicts).
 */
static void
smix2(uint8_t * B, size_t r, uint32_t N, uint32_t Nloop, escrypt_flags_t flags,
    void * V, uint32_t NROM, const escrypt_shared_t * shared, void * XY,
    void * S)
{
	const void * VROM = shared->shared1.aligned;
	uint32_t VROM_mask = shared->mask1 | 1;
	size_t s = 128 * r;
	__m128i * X = XY, * Y = (void *)((uintptr_t)(XY) + s);
	uint32_t * X32 = XY;
	uint32_t i, j;
	size_t k;

	if (Nloop == 0)
		return;

	/* X <-- B' */
	/* 3: V_i <-- X */
	for (k = 0; k < 2 * r; k++) {
		for (i = 0; i < 16; i++) {
			X32[k * 16 + i] =
			    le32dec(&B[(k * 16 + (i * 5 % 16)) * 4]);
		}
	}

	i = Nloop / 2;

	/* 7: j <-- Integerify(X) mod N */
	j = integerify(X, r) & (N - 1);

/*
 * Normally, NROM implies ESCRYPT_RW, but we check for these separately because
 * ESCRYPT_PARALLEL_SMIX resets ESCRYPT_RW for the smix2() calls operating on
 * the entire V.
 */
	if (NROM && (flags & ESCRYPT_RW)) {
		/* 6: for i = 0 to N - 1 do */
		for (i = 0; i < Nloop; i += 2) {
			__m128i * V_j = (void *)((uintptr_t)(V) + j * s);

			/* 8: X <-- H(X \xor V_j) */
			/* V_j <-- Xprev \xor V_j */
			/* j <-- Integerify(X) mod NROM */
			if (S)
				j = blockmix_sbox_xor_save(X, V_j, Y, S, r);
			else
				j = blockmix_salsa8_xor_save(X, V_j, Y, r);

			if (((i + 1) & VROM_mask) == 1) {
				const __m128i * VROM_j;

				j &= NROM - 1;
				VROM_j = (const void *)((uintptr_t)(VROM) + j * s);

				/* X <-- H(X \xor VROM_j) */
				/* 7: j <-- Integerify(X) mod N */
				if (S)
					j = blockmix_sbox_xor(Y, VROM_j, X, S, r, 1);
				else
					j = blockmix_salsa8_xor(Y, VROM_j, X, r, 1);
			} else {
				j &= N - 1;
				V_j = (void *)((uintptr_t)(V) + j * s);

				/* 8: X <-- H(X \xor V_j) */
				/* V_j <-- Xprev \xor V_j */
				/* j <-- Integerify(X) mod NROM */
				if (S)
					j = blockmix_sbox_xor_save(Y, V_j, X, S, r);
				else
					j = blockmix_salsa8_xor_save(Y, V_j, X, r);
			}
			j &= N - 1;
			V_j = (void *)((uintptr_t)(V) + j * s);
		}
	} else if (NROM) {
		/* 6: for i = 0 to N - 1 do */
		for (i = 0; i < Nloop; i += 2) {
			__m128i * V_j = (void *)((uintptr_t)(V) + j * s);
			const __m128i * VROM_j;

			/* 8: X <-- H(X \xor V_j) */
			/* V_j <-- Xprev \xor V_j */
			/* j <-- Integerify(X) mod NROM */
			if (S)
				j = blockmix_sbox_xor(X, V_j, Y, S, r, 0);
			else
				j = blockmix_salsa8_xor(X, V_j, Y, r, 0);

			if (((i + 1) & VROM_mask) == 1) {
				j &= NROM - 1;
				VROM_j = (const void *)((uintptr_t)(VROM) + j * s);
			} else {
				j &= N - 1;
				VROM_j = (const void *)((uintptr_t)(V) + j * s);
			}

			/* X <-- H(X \xor VROM_j) */
			/* 7: j <-- Integerify(X) mod N */
			if (S)
				j = blockmix_sbox_xor(Y, VROM_j, X, S, r, 1);
			else
				j = blockmix_salsa8_xor(Y, VROM_j, X, r, 1);
			j &= N - 1;
			V_j = (void *)((uintptr_t)(V) + j * s);
		}
	} else if (flags & ESCRYPT_RW) {
		/* 6: for i = 0 to N - 1 do */
		do {
			__m128i * V_j = (void *)((uintptr_t)(V) + j * s);

			/* 8: X <-- H(X \xor V_j) */
			/* V_j <-- Xprev \xor V_j */
			/* 7: j <-- Integerify(X) mod N */
			if (S)
				j = blockmix_sbox_xor_save(X, V_j, Y, S, r);
			else
				j = blockmix_salsa8_xor_save(X, V_j, Y, r);
			j &= N - 1;
			V_j = (void *)((uintptr_t)(V) + j * s);

			/* 8: X <-- H(X \xor V_j) */
			/* V_j <-- Xprev \xor V_j */
			/* 7: j <-- Integerify(X) mod N */
			if (S)
				j = blockmix_sbox_xor_save(Y, V_j, X, S, r);
			else
				j = blockmix_salsa8_xor_save(Y, V_j, X, r);
			j &= N - 1;
		} while (--i);
	} else {
		/* 6: for i = 0 to N - 1 do */
		do {
			const __m128i * V_j = (void *)((uintptr_t)(V) + j * s);

			/* 8: X <-- H(X \xor V_j) */
			/* 7: j <-- Integerify(X) mod N */
			if (S)
				j = blockmix_sbox_xor(X, V_j, Y, S, r, 0);
			else
				j = blockmix_salsa8_xor(X, V_j, Y, r, 0);
			j &= N - 1;
			V_j = (const void *)((uintptr_t)(V) + j * s);

			/* 8: X <-- H(X \xor V_j) */
			/* 7: j <-- Integerify(X) mod N */
			if (S)
				j = blockmix_sbox_xor(Y, V_j, X, S, r, 0);
			else
				j = blockmix_salsa8_xor(Y, V_j, X, r, 0);
			j &= N - 1;
		} while (--i);
	}

	/* 10: B' <-- X */
	X32 = XY;
	for (k = 0; k < 2 * r; k++) {
		for (i = 0; i < 16; i++) {
			le32enc(&B[(k * 16 + (i * 5 % 16)) * 4],
			    X32[k * 16 + i]);
		}
	}
}

/**
 * p2floor(x):
 * Largest power of 2 not greater than argument.
 */
static uint64_t
p2floor(uint64_t x)
{
	uint64_t y;
	while ((y = x & (x - 1)))
		x = y;
	return x;
}

/**
 * smix(B, r, N, p, flags, V, NROM, shared, XY, S):
 * Compute B = SMix_r(B, N).  The input B must be 128rp bytes in length; the
 * temporary storage V must be 128rN bytes in length; the temporary storage XY
 * must be 256r or 256rp bytes in length (the larger size is required with
 * OpenMP-enabled builds).  The value N must be a power of 2 greater than 1.
 * The array V must be aligned to a multiple of 64 bytes, and arrays B and
 * XY to a multiple of at least 16 bytes (aligning them to 64 bytes as well
 * saves cache lines and helps avoid false sharing in OpenMP-enabled builds
 * when p > 1, but it might also result in cache bank conflicts).
 */
static void
smix(uint8_t * B, size_t r, uint32_t N, uint32_t p, escrypt_flags_t flags,
    void * V, uint32_t NROM, const escrypt_shared_t * shared, void * XY,
    void * S)
{
	size_t s = 128 * r;
	uint32_t Nchunk = (N / p) & ~(uint32_t)1, Nloop;
	uint32_t i;

	Nloop = 0;
	if (flags & __ESCRYPT_INIT_SHARED)
		Nloop = Nchunk;
	else if (flags & ESCRYPT_RW)
		Nloop = (Nchunk / p) & ~(uint32_t)1;

#ifdef _OPENMP
#pragma omp parallel if (p > 1) default(none) private(i) shared(B, r, N, p, flags, V, NROM, shared, XY, S, s, Nchunk, Nloop)
	{
#pragma omp for
#endif
	for (i = 0; i < p; i++) {
		uint32_t Vchunk = i * Nchunk;
		uint8_t * Bp = &B[i * s];
		void * Vp = (void *)((uintptr_t)(V) + Vchunk * s);
#ifdef _OPENMP
		void * XYp = (void *)((uintptr_t)(XY) + i * (2 * s));
#else
		void * XYp = XY;
#endif
		uint32_t Np = (i < p - 1) ? Nchunk : (N - Vchunk);
		void * Sp = S ? (void *)((uintptr_t)(S) + i * S_SIZE_ALL) : S;
		if (Sp)
			smix1(Bp, 1, S_SIZE_ALL / 128,
			    flags & ~ESCRYPT_BLOCKMIX_SBOX,
			    Sp, NROM, shared, XYp, NULL);
		if (!(flags & __ESCRYPT_INIT_SHARED_2))
			smix1(Bp, r, Np, flags, Vp, NROM, shared, XYp, Sp);
		smix2(Bp, r, p2floor(Np), Nloop, flags, Vp,
		    NROM, shared, XYp, Sp);
	}

	if (Nchunk > Nloop) {
#ifdef _OPENMP
#pragma omp for
#endif
		for (i = 0; i < p; i++) {
			uint8_t * Bp = &B[i * s];
#ifdef _OPENMP
			void * XYp = (void *)((uintptr_t)(XY) + i * (2 * s));
#else
			void * XYp = XY;
#endif
			void * Sp = S ?
			    (void *)((uintptr_t)(S) + i * S_SIZE_ALL) : S;
			smix2(Bp, r, N, Nchunk - Nloop, flags & ~ESCRYPT_RW,
			    V, NROM, shared, XYp, Sp);
		}
	}
#ifdef _OPENMP
	}
#endif
}

/**
 * escrypt_kdf(shared, local, passwd, passwdlen, salt, saltlen,
 *     N, r, p, flags, buf, buflen):
 * Compute scrypt(passwd[0 .. passwdlen - 1], salt[0 .. saltlen - 1], N, r,
 * p, buflen) and write the result into buf.  The parameters r, p, and buflen
 * must satisfy r * p < 2^30 and buflen <= (2^32 - 1) * 32.  The parameter N
 * must be a power of 2 greater than 1.
 *
 * Return 0 on success; or -1 on error.
 */
int
escrypt_kdf(const escrypt_shared_t * shared, escrypt_local_t * local,
    const uint8_t * passwd, size_t passwdlen,
    const uint8_t * salt, size_t saltlen,
    uint64_t N, uint32_t r, uint32_t p, escrypt_flags_t flags,
    uint8_t * buf, size_t buflen)
{
	escrypt_region_t tmp;
	uint64_t NROM;
	size_t B_size, V_size, XY_size, need;
	uint8_t * B;
	uint32_t * V, * XY, * S;

	/* Sanity-check parameters. */
	if (flags & ~ESCRYPT_KNOWN_FLAGS) {
		errno = EINVAL;
		return -1;
	}
#if SIZE_MAX > UINT32_MAX
	if (buflen > (((uint64_t)(1) << 32) - 1) * 32) {
		errno = EFBIG;
		return -1;
	}
#endif
	if ((uint64_t)(r) * (uint64_t)(p) >= (1 << 30)) {
		errno = EFBIG;
		return -1;
	}
	if (N > UINT32_MAX) {
		errno = EFBIG;
		return -1;
	}
	if (((N & (N - 1)) != 0) || (N <= 7) || (r < 1) || (p < 1)) {
		errno = EINVAL;
		return -1;
	}
	if ((flags & ESCRYPT_PARALLEL_SMIX) && (N / p <= 7)) {
		errno = EINVAL;
		return -1;
	}
	if ((r > SIZE_MAX / 256 / p) ||
	    (N > SIZE_MAX / 128 / r)) {
		errno = ENOMEM;
		return -1;
	}
#ifdef _OPENMP
	if (!(flags & ESCRYPT_PARALLEL_SMIX) &&
	    (N > SIZE_MAX / 128 / (r * p))) {
		errno = ENOMEM;
		return -1;
	}
#endif
	if ((flags & ESCRYPT_BLOCKMIX_SBOX) &&
#ifndef _OPENMP
	    (flags & ESCRYPT_PARALLEL_SMIX) &&
#endif
	    p > SIZE_MAX / S_SIZE_ALL) {
		errno = ENOMEM;
		return -1;
	}

	NROM = 0;
	if (shared->shared1.aligned) {
		NROM = shared->shared1.aligned_size / ((size_t)128 * r);
		if (NROM > UINT32_MAX) {
			errno = EFBIG;
			return -1;
		}
		if (((NROM & (NROM - 1)) != 0) || (NROM <= 7) ||
		    !(flags & ESCRYPT_RW)) {
			errno = EINVAL;
			return -1;
		}
	}

	/* Allocate memory. */
	V = NULL;
	V_size = (size_t)128 * r * N;
#ifdef _OPENMP
	if (!(flags & ESCRYPT_PARALLEL_SMIX))
		V_size *= p;
#endif
	need = V_size;
	if (flags & __ESCRYPT_INIT_SHARED) {
		if (local->aligned_size < need) {
			if (local->base || local->aligned ||
			    local->base_size || local->aligned_size) {
				errno = EINVAL;
				return -1;
			}
			if (!alloc_region(local, need))
				return -1;
		}
		V = (uint32_t *)local->aligned;
		need = 0;
	}
	B_size = (size_t)128 * r * p;
	need += B_size;
	if (need < B_size) {
		errno = ENOMEM;
		return -1;
	}
	XY_size = (size_t)256 * r;
#ifdef _OPENMP
	XY_size *= p;
#endif
	need += XY_size;
	if (need < XY_size) {
		errno = ENOMEM;
		return -1;
	}
	if (flags & ESCRYPT_BLOCKMIX_SBOX) {
		size_t S_size = S_SIZE_ALL;
#ifdef _OPENMP
		S_size *= p;
#else
		if (flags & ESCRYPT_PARALLEL_SMIX)
			S_size *= p;
#endif
		need += S_size;
		if (need < S_size) {
			errno = ENOMEM;
			return -1;
		}
	}
	if (V) {
		if (!alloc_region(&tmp, need))
			return -1;
		B = (uint8_t *)tmp.aligned;
		XY = (uint32_t *)((uint8_t *)B + B_size);
	} else {
		if (local->aligned_size < need) {
			if (free_region(local))
				return -1;
			if (!alloc_region(local, need))
				return -1;
		}
		B = (uint8_t *)local->aligned;
		V = (uint32_t *)((uint8_t *)B + B_size);
		XY = (uint32_t *)((uint8_t *)V + V_size);
	}
	S = NULL;
	if (flags & ESCRYPT_BLOCKMIX_SBOX)
		S = (uint32_t *)((uint8_t *)XY + XY_size);

	/* 1: (B_0 ... B_{p-1}) <-- PBKDF2(P, S, 1, p * MFLen) */
	PBKDF2_SHA256(passwd, passwdlen, salt, saltlen, 1, B, B_size);

	if (p == 1 || (flags & ESCRYPT_PARALLEL_SMIX)) {
		smix(B, r, N, p, flags, V, NROM, shared, XY, S);
	} else {
		uint32_t i;

		/* 2: for i = 0 to p - 1 do */
#ifdef _OPENMP
#pragma omp parallel for default(none) private(i) shared(B, r, N, p, flags, V, NROM, shared, XY, S)
#endif
		for (i = 0; i < p; i++) {
			/* 3: B_i <-- MF(B_i, N) */
#ifdef _OPENMP
			smix(&B[(size_t)128 * r * i], r, N, 1, flags,
			    &V[(size_t)32 * r * i * N],
			    NROM, shared,
			    &XY[(size_t)64 * r * i],
			    S ? &S[S_SIZE_ALL / 4 * i] : S);
#else
			smix(&B[(size_t)128 * r * i], r, N, 1, flags, V,
			    NROM, shared, XY, S);
#endif
		}
	}

	/* 5: DK <-- PBKDF2(P, B, 1, dkLen) */
	PBKDF2_SHA256(passwd, passwdlen, B, B_size, 1, buf, buflen);

	if (flags & __ESCRYPT_INIT_SHARED) {
		if (free_region(&tmp))
			return -1;
	}

	/* Success! */
	return 0;
}
